// textsearch/python/csrc/text_search.h
//
// Copyright (c)  2023  Xiaomi Corporation (authors: Wei Kang)

#ifndef TEXTSEARCH_PYTHON_CSRC_TEXTSEARCH_H_
#define TEXTSEARCH_PYTHON_CSRC_TEXTSEARCH_H_

#include "pybind11/numpy.h"
#include "pybind11/pybind11.h"
#include "pybind11/stl.h"

namespace py = pybind11;

#endif // TEXTSEARCH_PYTHON_CSRC_TEXTSEARCH_H_
