// textsearch/python/csrc/utils.h
//
// Copyright (c)  2023  Xiaomi Corporation

#ifndef TEXTSEARCH_PYTHON_CSRC_UTILS_H_
#define TEXTSEARCH_PYTHON_CSRC_UTILS_H_

#include "textsearch/python/csrc/text_search.h"

namespace fasttextsearch {

void PybindUtils(py::module &m);

}

#endif // TEXTSEARCH_PYTHON_CSRC_UTILS_H_
